--[[
开放式云控接口
协议标准来自触摸精灵 http://ask.touchelf.net/docs/touchelfWebApi

本文件仅作为参考，请不要修改本文件
本文件会在重装、更新时被原版覆盖
--]]

local json = require('cjson.safe')
local lfs = require('lfs')
local noexecute = require('no_os_execute')

function lua_is_running()
    local c, _, r = xxtouch.post('/is_running')
    if c == 200 then
        return (json.decode(r) or {code = 0}).code ~= 0
    else
        return false
    end
end

function file_lastline(filename, last)
    last = tonumber(last) or 1
    last = math.floor(last)
    last = (last > 0 and last) or 1
    if last == 1 then
        local f = io.open(filename, 'r')
        if f then
            local s = f:read('*a')
            f:close()
            local t = s:split('\n')
            if t[#t] == '' or t[#t] == '\r' then
                table.remove(t, #t)
            end
            return t[#t]
        end
    else
        local f = io.open(filename, 'r')
        if f then
            local s = f:read('*a')
            f:close()
            local t = s:split('\n')
            if t[#t] == '' or t[#t] == '\r' then
                table.remove(t, #t)
            end
            if #t <= last then
                return table.concat(t, '\r\n')
            else
                return table.concat(t, '\r\n', #t - last, #t)
            end
        end
    end
    return ''
end

local function _read_conf()
	local conf
	local f = io.open(XXT_CONF_FILE_NAME, 'r')
	if f then
		conf = json.decode(f:read('*a')) or {}
		f:close()
	end
	conf = (type(conf) == 'table') and conf or {}
	return conf
end

local function _number_range(num, min, max, default)
    default = tonumber(default) or 0
    num = tonumber(num) or default
    num = (num <= max) and num or max
    num = (num >= min) and num or min
    return num
end

local proc_operations = dofile(XXT_BIN_PATH..'/module-proc-operations.lua')
local proc_value_operations = proc_operations.proc_value_operations
local proc_queue_operations = proc_operations.proc_queue_operations

local _message_switcher = {
    ['app/state'] = function(ws, msgtab)
        local wifi_ip = "127.0.0.1"
        for i,v in ipairs(device.ifaddrs()) do
            if #tostring(v[1])>2 and tostring(v[1]):sub(1, 2)=="en" then
                wifi_ip = v[2]
                if tostring(v[1])=="en0" then
                    break
                end
            end
        end
        local conf = _read_conf()
        local w, h = screen.size()
        msgtab.body = {
            app = {
                version = sys.zeversion(),
                license = 'X.X.T.',
            },
            script = {
                select = conf['current_script'],
                running = lua_is_running(),
            },
            system = {
                scrw = w,
                scrh = h,
                os = 'ios',
                name = device.name(),
                sn = device.serial_number(),
                ndid = string.format("%08X-%016X", sys.MGCopyAnswer('ChipID'), sys.MGCopyAnswer('UniqueChipID')),
                udid = device.udid(),
                version = sys.version(),
                ip = wifi_ip,
                battery = device.battery_level(),
                log = file_lastline(XXT_LOG_PATH..'/sys.log'),
            }
        }
        msgtab.error = ''
        return msgtab
    end,
    ['app/register'] = function(ws, msgtab)
        msgtab.error = '没有这项功能'
        return msgtab
    end,
    ['script/list'] = function(ws, msgtab)
        local retlist = {}
        for name in lfs.dir(XXT_SCRIPTS_PATH) do
            if string.match(name, '.+%.lua') or string.match(name, '.+%.xxt') or string.match(name, '.+%.xpp') --[[or string.match(name, '.+%.tep')]] then
                retlist[#retlist + 1] = name
            end
        end
        msgtab.body = retlist
        msgtab.error = ''
        return msgtab
    end,
    ['script/run'] = function(ws, msgtab)
        local c, _, r
        if type(msgtab.body) == 'table' and type(msgtab.body.name) == 'string' and msgtab.body.name ~= '' then
            c, _, r = xxtouch.post('/launch_script_file', '', json.encode{filename = XXT_SCRIPTS_PATH..'/'..msgtab.body.name})
        else
            c, _, r = xxtouch.post('/launch_script_file', '', '')
        end
        if c == 200 then
            local ret = json.decode(r)
            if type(ret) ~= 'table' then
                msgtab.error = ret
            else
                if ret.code ~= 0 then
                    msgtab.error = ret
                end
            end
        else
            msgtab.error = msgtab.type..': connection failed.'
        end
        return msgtab
    end,
    ['script/stop'] = function(ws, msgtab)
        local c, _, r = xxtouch.post('/recycle', '', '')
        if c == 200 then
            msgtab.error = ''
        else
            msgtab.error = msgtab.type..': connection failed.'
        end
        return msgtab
    end,
    ['script/encrypt'] = function(ws, msgtab)
        if type(msgtab.body) == 'table' and type(msgtab.body.name) == 'string' and msgtab.body.name ~= '' then
            local c, _, r = xxtouch.post('/encript_file', '', json.encode{
                in_file = XXT_SCRIPTS_PATH..'/'..msgtab.body.name,
                out_file = XXT_SCRIPTS_PATH..'/'..msgtab.body.name..'.xxt',
            })
            if c == 200 then
                local ret = json.decode(r)
                if type(ret) ~= 'table' then
                    msgtab.error = ret
                else
                    if ret.code ~= 0 then
                        msgtab.error = ret
                    end
                end
            else
                msgtab.error = msgtab.type..': connection failed.'
            end
        else
            msgtab.error = msgtab.type..': argument error.'
        end
        return msgtab
    end,
    ['script/get'] = function(ws, msgtab)
        msgtab.error = ''
        local path = XXT_SCRIPTS_PATH..'/'
        if type(msgtab.body) == 'table' and type(msgtab.body.name) == 'string' and msgtab.body.name ~= '' then
            path = path..msgtab.body.name
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        if lfs.attributes(path, 'mode') ~= 'directory' then
            local f, errmsg = io.open(path, 'r')
            if f then
                local s = f:read('*a')
                f:close()
                msgtab.body = s:base64_encode()
            else
                msgtab.error = msgtab.type..': can not read file `'..path..'` '..errmsg
            end
        else
            msgtab.error = msgtab.type..': can not read file `'..path..'`'
        end
        return msgtab
    end,
    ['script/put'] = function(ws, msgtab)
        msgtab.error = ''
        local path = XXT_SCRIPTS_PATH..'/'
        if type(msgtab.body) == 'table' and type(msgtab.body.name) == 'string' and type(msgtab.body.data) == 'string' and msgtab.body.name ~= '' then
            path = path..msgtab.body.name
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        if lfs.attributes(path, 'mode') == 'directory' then
            msgtab.error = msgtab.type..': `'..path..'` is a directory.'
            return msgtab
        end
        local f, errmsg = io.open(path, 'w')
        if f then
            local data = msgtab.body.data:base64_decode()
            if type(data) == 'string' then
                f:write(data)
            else
                f:write(msgtab.body.data)
            end
            f:close()
            sys.lchown(path, 501, 501)
        else
            msgtab.error = msgtab.type..': can not write file `'..path..'` '..errmsg
        end
        return msgtab
    end,
    ['script/delete'] = function(ws, msgtab)
        msgtab.error = ''
        local path = XXT_SCRIPTS_PATH..'/'
        if type(msgtab.body) == 'table' and type(msgtab.body.name) == 'string' and msgtab.body.name ~= '' then
            path = path..msgtab.body.name
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        local pathinfo = lfs.attributes(path)
        if type(pathinfo) ~= 'table' then
            msgtab.error = msgtab.type..': no such file `'..path..'`'
            return msgtab
        end
        if pathinfo.mode == 'directory' then
            msgtab.error = msgtab.type..': `'..path..'` is a directory.'
            return msgtab
        end
        noexecute.rm_rf(path)
        return msgtab
    end,
    ['system/log/get'] = function(ws, msgtab)
        msgtab.error = ''
        local last = 0
        if type(msgtab.body) == 'table' and type(msgtab.body.last) == 'number' and msgtab.body.last > 0  then
            last = msgtab.body.last
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        local logfilepath = XXT_LOG_PATH..'/sys.log'
        msgtab.body = file_lastline(logfilepath, last)
        return msgtab
    end,
    ['system/log/delete'] = function(ws, msgtab)
        msgtab.error = ''
        noexecute.rm_rf(XXT_LOG_PATH..'/sys.log')
        return msgtab
    end,
    ['system/reboot'] = function(ws, msgtab)
        msgtab.error = ''
        xxtouch.post('/reboot2', '', '')
        return msgtab
    end,
    ['system/respring'] = function(ws, msgtab)
        msgtab.error = ''
        xxtouch.post('/respring', '', '')
        return msgtab
    end,
    ['screen/snapshot'] = function(ws, msgtab)
        msgtab.error = ''
        local scale = 100
        local format = 'png'
        if type(msgtab.body) == 'table' then
            scale = _number_range(msgtab.body.scale, 1, 100, scale)
            format = tostring(msgtab.body.format)
        end
        local function image_data_with_format(img, format)
            if format == 'jpg' or format == 'jpeg' then
                return img:jpeg_data()
            else
                return img:png_data()
            end
        end
        if scale == 100 then
            msgtab.body = image_data_with_format(screen.image(), format):base64_encode()
        else
            local zoom = scale / 100
            local img = screen.image()
            local w, h = img:size()
            require('image.cv')
            img = img:cv_resize(math.floor(w * zoom), math.floor(h * zoom))
            msgtab.body = image_data_with_format(img, format):base64_encode()
        end
        return msgtab
    end,
    ['file/list'] = function(ws, msgtab)
        msgtab.error = ''
        local path = XXT_HOME_PATH..'/'
        if type(msgtab.body) == 'table' and type(msgtab.body.path) == 'string' and msgtab.body.path ~= '' then
            path = path..msgtab.body.path
        end
        if lfs.attributes(path, 'mode') == 'directory' then
            local retlist = {}
            setmetatable(retlist, json.array_mt)
            for name in lfs.dir(path) do
                if name~='.' and name~='..' then
                    local finfo = lfs.attributes(path..'/'..name)
                    if finfo then
                        if finfo.mode == 'directory' then
                            retlist[#retlist + 1] = {name = name, type = 'dir'}
                        else
                            retlist[#retlist + 1] = {name = name, type = 'file'}
                        end
                    end
                end
            end
            msgtab.body = retlist
        else
            msgtab.error = msgtab.type..': `'..path..'` is not a directory'
        end
        return msgtab
    end,
    ['file/get'] = function(ws, msgtab)
        msgtab.error = ''
        local path = XXT_HOME_PATH..'/'
        if type(msgtab.body) == 'table' and type(msgtab.body.path) == 'string' and msgtab.body.path ~= '' then
            path = path..msgtab.body.path
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        if lfs.attributes(path, 'mode') ~= 'directory' then
            local f, errmsg = io.open(path, 'r')
            if f then
                local s = f:read('*a')
                f:close()
                msgtab.body = s:base64_encode()
            else
                msgtab.error = msgtab.type..': can not read file `'..path..'` '..errmsg
            end
        else
            msgtab.error = msgtab.type..': can not read file `'..path..'`'
        end
        return msgtab
    end,
    ['file/put'] = function(ws, msgtab)
        msgtab.error = ''
        local path = XXT_HOME_PATH..'/'
        if type(msgtab.body) == 'table' and type(msgtab.body.path) == 'string' then
            path = path..msgtab.body.path
        else
            msgtab.error = msgtab.type..': argument error.'
            msgtab.body.data = nil
            return msgtab
        end
        if msgtab.body.directory == true then
            sys.mkdir_p(path)
            sys.lchown(path, 501, 501)
            msgtab.body.data = nil
            return msgtab
        end
        if type(msgtab.body.data) ~= 'string' or msgtab.body.path == '' then
            msgtab.error = msgtab.type..': argument error.'
            msgtab.body.data = nil
            return msgtab
        end
        if lfs.attributes(path, 'mode') == 'directory' then
            msgtab.error = msgtab.type..': `'..path..'` is a directory.'
            msgtab.body.data = nil
            return msgtab
        end
        local f, errmsg = io.open(path, 'w')
        if f then
            local data = msgtab.body.data:base64_decode()
            if type(data) == 'string' then
                f:write(data)
            else
                f:write(msgtab.body.data)
            end
            f:close()
            sys.lchown(path, 501, 501)
        else
            msgtab.error = msgtab.type..': can not write file `'..path..'` '..errmsg
        end
        msgtab.body.data = nil
        return msgtab
    end,
    ['file/delete'] = function(ws, msgtab)
        msgtab.error = ''
        local path = XXT_HOME_PATH..'/'
        if type(msgtab.body) == 'table' and type(msgtab.body.path) == 'string' and msgtab.body.path ~= '' and msgtab.body.path:gsub('/', '') ~= '' then
            path = path..msgtab.body.path
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        if lfs.attributes(path, 'mode') then
            noexecute.rm_rf(path)
        end
        return msgtab
    end,
    ['touch/down'] = function(ws, msgtab)
        msgtab.error = ''
        local body = type(msgtab.body) == 'table' and msgtab.body or {}
        local finger = math.floor(tonumber(body.finger) or 1)
        local x = math.floor(tonumber(body.x) or -1) or -1
        local y = math.floor(tonumber(body.y) or -1) or -1
        if finger >= 0 and finger <= 29 and x >= 0 and y >= 0 then
            touch.down(finger, x, y)
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        return msgtab
    end,
    ['touch/move'] = function(ws, msgtab)
        msgtab.error = ''
        local body = type(msgtab.body) == 'table' and msgtab.body or {}
        local finger = math.floor(tonumber(body.finger) or 1)
        local x = math.floor(tonumber(body.x) or -1) or -1
        local y = math.floor(tonumber(body.y) or -1) or -1
        if finger >= 0 and finger <= 29 and x >= 0 and y >= 0 then
            touch.move(finger, x, y)
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        return msgtab
    end,
    ['touch/up'] = function(ws, msgtab)
        msgtab.error = ''
        local body = type(msgtab.body) == 'table' and msgtab.body or {}
        local finger = math.floor(tonumber(body.finger) or 1)
        if finger >= 0 and finger <= 29 then
            touch.up(finger)
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        return msgtab
    end,
    ['touch/tap'] = function(ws, msgtab)
        msgtab.error = ''
        local body = type(msgtab.body) == 'table' and msgtab.body or {}
        local finger = math.floor(tonumber(body.finger) or 1)
        local x = math.floor(tonumber(body.x) or -1) or -1
        local y = math.floor(tonumber(body.y) or -1) or -1
        if finger >= 0 and finger <= 29 and x >= 0 and y >= 0 then
            touch.down(finger, x, y)
            touch.up(finger)
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        return msgtab
    end,
    ['key/down'] = function(ws, msgtab)
        msgtab.error = ''
        local body = type(msgtab.body) == 'table' and msgtab.body or {}
        local keycode = type(body.code) == 'string' and body.code or nil
        if keycode then
            if keycode:upper() == "HOME" then
                keycode = "HOMEBUTTON"
            end
            local ok, err = pcall(key.down, keycode)
            if not ok then
                msgtab.error = msgtab.type..': '..err
                return msgtab
            end
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        return msgtab
    end,
    ['key/up'] = function(ws, msgtab)
        msgtab.error = ''
        local body = type(msgtab.body) == 'table' and msgtab.body or {}
        local keycode = type(body.code) == 'string' and body.code or nil
        if keycode then
            if keycode:upper() == "HOME" then
                keycode = "HOMEBUTTON"
            end
            local ok, err = pcall(key.up, keycode)
            if not ok then
                msgtab.error = msgtab.type..': '..err
                return msgtab
            end
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        return msgtab
    end,
    ['key/press'] = function(ws, msgtab)
        msgtab.error = ''
        local body = type(msgtab.body) == 'table' and msgtab.body or {}
        local keycode = type(body.code) == 'string' and body.code or nil
        if keycode then
            if keycode:upper() == "HOME" then
                keycode = "HOMEBUTTON"
            end
            local ok, err = pcall(key.press, keycode)
            if not ok then
                msgtab.error = msgtab.type..': '..err
                return msgtab
            end
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        return msgtab
    end,
    ['pasteboard/read'] = function(ws, msgtab)
        msgtab.error = ''
        local data = pasteboard.read("public.plain-text")
        if data ~= "" then
            msgtab.body = {
                uti = "public.plain-text",
                data = data,
            }
        else
            data = pasteboard.read('public.image')
            if data ~= "" then
                msgtab.body = {
                    uti = "public.image",
                    data = data:base64_encode(),
                }
            else
                msgtab.error = msgtab.type..': supports copying plain text or plain images from devices only.'
                return msgtab
            end
        end
        return msgtab
    end,
    ['pasteboard/write'] = function(ws, msgtab)
        msgtab.error = ''
        local body = type(msgtab.body) == 'table' and msgtab.body or {}
        local uti = type(body.uti) == 'string' and body.uti or "public.plain-text"
        local data = type(body.data) == 'string' and body.data or nil
        if data then
            if uti == "public.plain-text" then
                pasteboard.write(data, uti)
            elseif uti == "public.png" or uti == "public.jpeg" or uti == "public.image" then
                local imgdata = data:base64_decode()
                pasteboard.write(imgdata, uti)
            else
                msgtab.error = msgtab.type..': transfer of clipboard contents other than plain text and plain images is not supported.'
                return msgtab
            end
        else
            msgtab.error = msgtab.type..': argument error.'
            return msgtab
        end
        return msgtab
    end,
    ['path/jbroot'] = function(ws, msgtab)
        if type(msgtab.body) == 'table' and type(msgtab.body.path) == 'string' and msgtab.body.path ~= '' and msgtab.body.path:gsub('/', '') ~= '' then
            msgtab.body = jbroot(msgtab.body.path)
        else
            msgtab.body = jbroot('/')
        end
        return msgtab
    end,
    ['path/rootfs'] = function(ws, msgtab)
        if type(msgtab.body) == 'table' and type(msgtab.body.path) == 'string' and msgtab.body.path ~= '' and msgtab.body.path:gsub('/', '') ~= '' then
            msgtab.body = rootfs(msgtab.body.path)
        else
            msgtab.body = rootfs('/')
        end
        return msgtab
    end,
}

local function proc_value_handler(ws, msgtab)
    local body = msgtab.body
    if type(body) ~= 'table' or type(body.key) ~= 'string' then
        msgtab.error = msgtab.type..': argument error.'
        return msgtab
    end
    local operation_name = (msgtab.type:split('/'))[2]
    local operation = proc_value_operations[operation_name]
    if type(operation) ~= 'function' then
        msgtab.error = msgtab.type..': operation not found.'
        return msgtab
    end
    local ret, err = operation(body)
    if type(ret) == 'table' then
        msgtab.body = ret
    else
        msgtab.error = msgtab.type..': operation failed: '..(err or 'unknown error')
    end
    return msgtab
end

local function proc_queue_handler(ws, msgtab)
    local body = msgtab.body
    if type(body) ~= 'table' or type(body.key) ~= 'string' then
        msgtab.error = msgtab.type..': argument error.'
        return msgtab
    end
    local operation_name = (msgtab.type:split('/'))[2]
    local operation = proc_queue_operations[operation_name]
    if type(operation) ~= 'function' then
        msgtab.error = msgtab.type..': operation not found.'
        return msgtab
    end
    local ret, err = operation(body)
    if type(ret) == 'table' then
        msgtab.body = ret
    else
        msgtab.error = msgtab.type..': operation failed: '..(err or 'unknown error')
    end
    return msgtab
end

_message_switcher['proc-value/put'] = proc_value_handler
_message_switcher['proc-value/get'] = proc_value_handler
_message_switcher['proc-queue/push'] = proc_queue_handler
_message_switcher['proc-queue/push-back'] = proc_queue_handler
_message_switcher['proc-queue/push-front'] = proc_queue_handler
_message_switcher['proc-queue/pop'] = proc_queue_handler
_message_switcher['proc-queue/pop-front'] = proc_queue_handler
_message_switcher['proc-queue/pop-back'] = proc_queue_handler
_message_switcher['proc-queue/count-value'] = proc_queue_handler
_message_switcher['proc-queue/pop-value'] = proc_queue_handler
_message_switcher['proc-queue/read'] = proc_queue_handler
_message_switcher['proc-queue/clear'] = proc_queue_handler

function _table_message_router(ws, msgtab)
    local handler = _message_switcher[msgtab.type]
    if type(handler) ~= 'function' then
        return nil
    end
    local success, ret = pcall(handler, ws, msgtab)
    if success then
        local json_ret = json.encode(ret)
        if type(json_ret) == 'string' then
            return json_ret
        else
            return nil
            -- return error('open-cloud-control-client.lua: _table_message_router: json_err_ret invalid')
        end
    else
        msgtab.error = ret
        local json_err_ret = json.encode(msgtab)
        if type(json_err_ret) == 'string' then
            return json_err_ret
        else
            return nil
            -- return error('open-cloud-control-client.lua: _table_message_router: json_err_ret invalid')
        end
    end
end

function _message_router(ws, msg)
    local msgtab = json.decode(msg)
    if type(msgtab) == 'table' then
        if type(msgtab.type) == 'string' then
            return _table_message_router(ws, msgtab)
        else
            return error('open-cloud-control-client.lua: _message_router: message.type invalid')
        end
    end
    return error('open-cloud-control-client.lua: _message_router: message invalid')
end

local conf = _read_conf()

if type(conf) == 'table' and type(conf.open_cloud_control) == 'table' and type(conf.open_cloud_control.enable) == 'boolean' and type(conf.open_cloud_control.address) == 'string' and conf.open_cloud_control.enable then
    local ev = require('ev')
    local frame = require'websocket.frame'
    local loop = ev.Loop.default

    local ws_connect
    local ws_client
    local lifetime

    local function ws_on_message(ws, message, opcode)
        lifetime = 31
        if opcode == frame.TEXT then
            local success, ret = pcall(_message_router, ws, message)
            if success then
                if type(ret) == 'string' then
                    ws:send(ret)
                else
                    ws:send(message) -- 成功无返回值则 echo
                end
            else
                sys.log(ret)
            end
        end
    end

    function ws_connect()
        ws_client = require('websocket.client').ev()
        ws_client:connect(conf.open_cloud_control.address, 'echo-protocol')
        lifetime = 31

        ws_client:on_open(function(ws)
            lifetime = 31
            sys.log('open-cloud-control-client.lua: connected.')
            local success, ret = pcall(_message_router, ws, json.encode{
                type = "app/state",
            })
            sys.log(success, ret)
            if success then
                ws:send(ret)
            end
        end)

        ws_client:on_message(ws_on_message)
        -- ws_client:on_close(ws_connect)
    end

    ws_connect()

    ev.Timer.new(function()
        if lifetime <= 0 then
            sys.log('open-cloud-control-client.lua: timeout, reconnecting...')
            ws_client:close()
            ws_connect()
        end
        lifetime = lifetime - 1
    end, 1, 1):start(loop)

    dispatch_async("concurrent", function ()
        exit_callback_handle = notification_center_register_callback({
            center = "darwin",
            name = "xxtouch.open-cloud-control-client-service/exit",
        },
        function()
            exit_callback_handle:release()
            os.exit(0)
        end)
        CFRunLoopRunWithAutoreleasePool()
    end)

    loop:loop()
end
